#include "Twain.h"

enum{
	TWCC_UNUSABLETWAIN = TWCC_CUSTOMBASE,
	TWCC_DSM_IS_ALREADYOPEN,
	TWCC_DS_IS_ALREADYOPEN,
	TWCC_DS_IS_ALREADYENABLE,
	TWCC_DSM_IS_NOTOPEN,
	TWCC_DS_IS_NOTOPEN,
	TWCC_DS_IS_NOTENABLE,
	TWCC_PROGRAM_ERROR
};

typedef void (*ScanCallBack)(TW_UINT32 twBitmap);

HINSTANCE		g_hDSMDll = NULL;
DSMENTRYPROC	MyDSM_Entry = NULL;
TW_IDENTITY		g_appID;
TW_IDENTITY		g_DSID;
bool			g_bOpenDSM = false;
bool			g_bOpenDS = false;
bool			g_bEnableDS = false;
ScanCallBack	g_pfnCallBack = NULL;
HGLOBAL			g_hDIB = NULL;
LPVOID			g_pBitmap = NULL;

bool InitTWAIN();
BOOL ExitTWAIN();
TW_UINT16 OpenDSM(HWND hWnd);
TW_UINT16 CloseDSM(HWND hWnd);
TW_UINT16 SelectDS();
TW_UINT16 OpenDS();
TW_UINT16 CloseDS();
TW_UINT16 EnableDS(HWND hWnd);
TW_UINT16 DisableDS(HWND hWnd);
TW_UINT16 SelectScaner(HWND hWnd);
TW_UINT16 Scan(HWND hWnd, ScanCallBack pfnCallBack);
bool TwainTranslateMessage(HWND hWnd, MSG* pMsg);
TW_UINT16 NativeXfer();
void ScanBitmap(TW_UINT32 twBitmap);
void FreeScanData();

bool InitTWAIN()
{
	ZeroMemory(&g_DSID, sizeof(TW_IDENTITY));
	ZeroMemory(&g_appID, sizeof(TW_IDENTITY));
	
	if(!MyDSM_Entry && !g_hDSMDll){
		if((g_hDSMDll = LoadLibrary("TWAIN_32.DLL")) != NULL){
			if((MyDSM_Entry = (DSMENTRYPROC)GetProcAddress(g_hDSMDll, "DSM_Entry")) == NULL){
				FreeLibrary(g_hDSMDll);
				return false;
			}else{
				g_appID.Id = 0;
				g_appID.Version.MajorNum = 1;
				g_appID.Version.MinorNum = 0;
				g_appID.Version.Language = TWLG_JAPANESE;
				g_appID.Version.Country  = TWCY_JAPAN;
				strcpy (g_appID.Version.Info,  "\tg");
				strcpy (g_appID.ProductName,   "\tg");
				g_appID.ProtocolMajor = TWON_PROTOCOLMAJOR;
				g_appID.ProtocolMinor = TWON_PROTOCOLMINOR;
				g_appID.SupportedGroups =  DG_IMAGE | DG_CONTROL;
				strcpy (g_appID.Manufacturer,  "Җ");
				strcpy (g_appID.ProductFamily, "\tg");
				return true;
			}
		}
	}
	return false;
}// InitTWAIN

BOOL ExitTWAIN()
{
	FreeScanData();
	if(g_hDSMDll){
		BOOL bRet = FreeLibrary(g_hDSMDll);
		g_hDSMDll = NULL;
		MyDSM_Entry = NULL;
		return bRet;
	}
	return FALSE;
}// ExitTWAIN

TW_UINT16 OpenDSM(HWND hWnd)
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	if(g_bOpenDSM)		return TWCC_DSM_IS_ALREADYOPEN;
	TW_UINT16 twRC = MyDSM_Entry(&g_appID, NULL, DG_CONTROL, DAT_PARENT, MSG_OPENDSM, (TW_MEMREF)&hWnd);
	g_bOpenDSM = (twRC == TWRC_SUCCESS);
	return twRC;
}// OpenDSM

TW_UINT16 CloseDSM(HWND hWnd)
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	if(!g_bOpenDSM)		return TWCC_DSM_IS_NOTOPEN;
	g_bOpenDSM = FALSE;
	return MyDSM_Entry(&g_appID, NULL, DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, (TW_MEMREF)&hWnd);
}// CloseDSM

TW_UINT16 SelectDS()
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	TW_IDENTITY newDSID;
	ZeroMemory(&newDSID, sizeof(TW_IDENTITY));
	TW_UINT16 twRC;
	twRC = MyDSM_Entry(&g_appID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_GETDEFAULT, (TW_MEMREF)&newDSID);
	twRC = MyDSM_Entry(&g_appID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, (TW_MEMREF)&newDSID);
	if(twRC == TWRC_SUCCESS){
		g_DSID = newDSID; 
	}
	return twRC;
}// SelectDS

TW_UINT16 OpenDS()
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	if(g_DSID.Id == 0){
		strcpy(g_DSID.ProductName, "\0");
	}
	if (g_bOpenDS)		return TWCC_DS_IS_ALREADYOPEN;
	TW_UINT16 twRC = MyDSM_Entry(&g_appID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, (TW_MEMREF)&g_DSID);
	if(twRC == TWRC_SUCCESS){
		g_bOpenDS = true;
	}
	return twRC;
}// OpenDS

TW_UINT16 CloseDS()
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	g_bOpenDS = false;
	return MyDSM_Entry(&g_appID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, (TW_MEMREF)&g_DSID);
}// CloseDS

TW_UINT16 EnableDS(HWND hWnd)
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	if(g_bEnableDS)		return TWCC_DS_IS_ALREADYENABLE;
	TW_USERINTERFACE ui;
	ui.ShowUI = TRUE;
	ui.ModalUI = TRUE;
	ui.hParent = hWnd;
	TW_UINT16 twRC = MyDSM_Entry(&g_appID, &g_DSID, DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, (TW_MEMREF)&ui);
	if(twRC == TWRC_SUCCESS){
		EnableWindow(hWnd, FALSE);
		g_bEnableDS = true;
	}
	return twRC;
}// EnableDS

TW_UINT16 DisableDS(HWND hWnd)
{
	if(!MyDSM_Entry)	return TWCC_UNUSABLETWAIN;
	if(!g_bEnableDS)	return TWCC_DS_IS_NOTENABLE;
	TW_USERINTERFACE ui;
	TW_UINT16 twRC = MyDSM_Entry(&g_appID, &g_DSID, DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, (TW_MEMREF)&ui);
	g_bEnableDS = false;
	EnableWindow(hWnd, TRUE);
	return twRC;
}// DisableDS

TW_UINT16 SelectScaner(HWND hWnd)
{
	TW_UINT16 twRC = TWCC_UNUSABLETWAIN;
	if(MyDSM_Entry){
		if((twRC = OpenDSM(hWnd)) == TWRC_SUCCESS){
			twRC = SelectDS();
			CloseDSM(hWnd);
		}
	}
	return twRC;
}// SelectScaner

TW_UINT16 Scan(HWND hWnd, ScanCallBack pfnCallBack)
{
	g_pfnCallBack = NULL;
	TW_UINT16 twRC = TWCC_UNUSABLETWAIN;
	if(MyDSM_Entry){
		if((twRC = OpenDSM(hWnd)) == TWRC_SUCCESS){
			if((twRC = OpenDS()) == TWRC_SUCCESS){
				g_pfnCallBack = pfnCallBack;
				twRC = EnableDS(hWnd);
				return twRC;
			}
			CloseDS();
		}
		CloseDSM(hWnd);
	}
	return twRC;
}// Scan

bool TwainTranslateMessage(HWND hWnd, MSG* pMsg)
{
	TW_UINT16 twRC = TWRC_NOTDSEVENT;
	TW_EVENT twEvent;
	memset(&twEvent, 0, sizeof(TW_EVENT));
	if(g_bEnableDS){
		twEvent.pEvent = (TW_MEMREF)pMsg;
		twEvent.TWMessage = MSG_NULL;
		twRC = MyDSM_Entry(&g_appID, &g_DSID, DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT, (TW_MEMREF)&twEvent);
		if(twRC == TWRC_DSEVENT){
			switch(twEvent.TWMessage){
			case MSG_XFERREADY:
				switch(NativeXfer()){
				case TWRC_FAILURE:
					MessageBox(hWnd, "Transfer failure", "XLiǂݎ", MB_ICONWARNING);
					break;
				case TWRC_XFERDONE:
					break;
				default:
				case TWRC_CANCEL:
					break;
				}
				DisableDS(hWnd);
				CloseDS();
				CloseDSM(hWnd);
				break;
			case MSG_CLOSEDSREQ:
				DisableDS(hWnd);
				CloseDS();
				CloseDSM(hWnd);
				break;
			case MSG_CLOSEDSOK:
				DisableDS(hWnd);
				CloseDS();
				CloseDSM(hWnd);
				break;
			default:
				break;
			}
			return true;
		}
	}
	return false;
}// TwainTranslateMessage

TW_UINT16 NativeXfer()
{
	TW_UINT32 twBitmap;
	if(!g_pfnCallBack)	return TWCC_PROGRAM_ERROR;
	TW_INT16 twRC = MyDSM_Entry(&g_appID, &g_DSID, DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET, (TW_MEMREF)&twBitmap);
	switch(twRC){
	case TWRC_XFERDONE:
		(*g_pfnCallBack)(twBitmap);
		break;
	case TWRC_CANCEL:
		break;
	case TWRC_FAILURE:
		break;
	}
	return twRC;
}// NativeXfer

void ScanBitmap(TW_UINT32 twBitmap)
{
	FreeScanData();
	g_hDIB = (HGLOBAL) twBitmap;
	g_pBitmap = GlobalLock(g_hDIB);
}// ScanBitmap

void FreeScanData()
{
	GlobalUnlock(g_hDIB);
	GlobalFree(g_hDIB);
	g_hDIB = NULL;
	g_pBitmap = NULL;
}// FreeScanData
